/* ctcompare: Compare two C tokens files for similarities.
 * Copyright (c) Warren Toomey, under the GPL license.
 *
 * $Revision: 1.27 $
 */
#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <unistd.h>
#include "tokens.h"
#define ENDLIST 0xff
#undef DEBUG
#define OPT_OVERLAP		/* Turn on to enable overlap optimisations */
				/* It seems to double the run-time, but does */
				/* prevent duplicate output. */
#define ISOMORPH_CODE		/* Turn this on to reduce false-positives by */
				/* applying some isomorphic comparison code.*/

/* For each token, we have a node
 * which contains the type of
 * token, and a specific numeric
 * id for 5 token types: chars,
 * stringlits, numbers, identifiers and labels.
 */
struct token {
  unsigned short id;
  unsigned char type;
};

/*
 * For each file in a ctf file, we have a
 * struct sourcefile node which points
 * at the filename and the first token.
 * We also record the number of tokens
 * in the file.
 */
struct sourcefile {
  unsigned char *filename;
  struct token  *token;
  int numtokens;
};

struct sourcefile *Srctree, *Dsttree;
unsigned char *Srctoklist, *Dsttoklist;
int threshold = 20;		/* Threshold above which we report a run */
int showtokens = 0;		/* If 1, print out tokens in a run */
int crosscompare = 0;		/* If 1, print out actual code in a run */
int progress = 0;		/* Print filenames to stderr as we process */
int skip_cppstuff = 1;		/* Skip #defines, #includes etc. */

/* Array used to print out a run of tokens */
unsigned char *tokstring[] = {"ERR ","ERR ","ERR ","ERR ","ERR ","ERR ","ERR ",
     "ERR ", "ERR ", "ERR ", "\n", "ERR ", "ERR ", ">>= ", "ERR ", "ERR ",
     "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ",
     "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "ERR ", "/= ", "! ",
     "\"string", "->", "++ ", "% ", "& ", "'c","( ",") ","* ","+ ",", ","- ",
     ". ", "/ ", "-- ", "&& ", "|| ", "++= ", "%= ", "-= ", "&= ", "*= ", "|= ",
     "NUM", ": ", "; ", "< ", "= ", "> ", "? ", "!= ", "<= ", "case ", "char ",
     "const ", "continue ", "default ", "do ", "...", "double ", "else ",
     "enum ", "extern ", "float ", "for ", "goto ", "if ", "int ", "long ",
     "register ", "return", "short ", "signed ", "sizeof ","static ", "struct ",
     "switch ", "[ ", "\\", "] ", "^ ", "id", "label", "typedef ", "union ",
     "unsigned ", "void ", "volatile ", "while ", "#define ", "#elif ",
     "#else ", "#endif ", "#error ", "#ifdef ", "#if ", "#ifndef ", "#include ",
     "#line ", "#pragma ", "#undef ", "#warning ", "~= ", "== ", "break ",">= ",
     "<< ", ">> ", "<<= ", "{ ", "| ", "} ", "~ ", "ERR " };


/*
 * Load the ctf file into a char array.
 * Returns the array pointer or NULL if
 * something goes wrong. The 5-octet ctf
 * header is stripped.
 */
unsigned char *
load_ctffile(char *filename)
{
  FILE *zin;
  struct stat sb;
  int err, cnt;
  int size;
  unsigned char *ctfarray, *cptr;

  /* Get the file's size */
  err = stat(filename, &sb);
  if (err < 0) {
    fprintf(stderr, "%s doesn't exist\n", filename);
    return (NULL);
  }
  size = sb.st_size;

  /* Malloc an array for it. Yes I know I don't need all of size */
  ctfarray = (char *)malloc(size);
  if (ctfarray == NULL) {
    fprintf(stderr, "Can't malloc!\n");
    return (NULL);
  }
  /* Open the file */
  zin = fopen(filename, "r");
  if (zin == NULL) {
    fprintf(stderr, "Can't open %s\n", filename);
    return (NULL);
  }
  /* Check the ctf header */
  if ((fgetc(zin) != 'c') || (fgetc(zin) != 't') || (fgetc(zin) != 'f') ||
      (fgetc(zin) != 0x1) || (fgetc(zin) != 0x1)) {
    fprintf(stderr, "File %s is not a ctf file\n", filename);
    exit(1);
  }
  /* Read the rest of the file in */
  cptr = ctfarray; size-= 5;
  while (1) {
    cnt = fread(cptr, 1, size, zin);
    if (cnt < 1)
      break;
    size -= cnt;
    cptr+= cnt;
  }
  *cptr = ENDLIST;		/* Ensure the end of the array is marked */
  fclose(zin);
  return (ctfarray);
}


/*
 * Given a token array, build an array of struct sourcefile
 * nodes, each of which has an array of struct token nodes.
 * The last node in the struct sourcefile array will have
 * NULL pointers. If an error occurs, return NULL.
 */

struct sourcefile *
build_srcarray(unsigned char *tokarray)
{
  struct sourcefile *S = NULL;
  int value;
  int numfiles=0;
  int fileindex=0;
  int tokindex=0;
  int inside_cppstuff=0;
  unsigned char *tokptr;


  /* First step: count the number of files in the ctf file */
  for (tokptr= tokarray; *tokptr != ENDLIST; tokptr++) {
    switch(*tokptr) {
      case FILENAME:
		numfiles++; tokptr++;
		tokptr+= strlen(tokptr); break;
      case IDENTIFIER:
      case INTVAL:
      case LABEL:
      case CHARCONST:
      case STRINGLIT:
		tokptr+=2;
    }
  }

  /* Allocate an array of struct sourcefile nodes to hold the details */
  numfiles++;
  S= (struct sourcefile *) calloc(numfiles, sizeof(struct sourcefile));
  if (S == NULL) {
    fprintf(stderr, "calloc failure\n");
    return (NULL);
  }

  /* NULL out the last node */
  numfiles--;
  S[numfiles].filename=NULL;
  S[numfiles].token=NULL;
  S[numfiles].numtokens=0;

  /* Second step: point the filenames into the buffer
   * and count the number of tokens in each file.
   * We start with fileindex==-1 as we get incremented
   * to 0 by the first FILENAME token.
   */
  for (fileindex=-1, tokptr= tokarray; *tokptr != ENDLIST; tokptr++) {
    switch(*tokptr) {
      case FILENAME:
		fileindex++; tokptr++;
		S[fileindex].filename = tokptr;
		S[fileindex].token=NULL;
		S[fileindex].numtokens=0;
		tokptr+= strlen(tokptr); break;
      case IDENTIFIER:
      case INTVAL:
      case LABEL:
      case CHARCONST:
      case STRINGLIT:
		tokptr+=2;
      default:
		S[fileindex].numtokens++;
    }
  }

  /* Third step: allocate the token nodes for
   * each file and fill them in.
   */
  for (fileindex=0; fileindex<numfiles; fileindex++) {

    /* Malloc the token list */
    S[fileindex].token=
	(struct token *)malloc((S[fileindex].numtokens+1)*sizeof(struct token));
    if (S[fileindex].token == NULL) {
      fprintf(stderr, "malloc failure\n");
      return (NULL);
    }

    /* Now walk the tokens for this file, and build up each token node */
    /* tokptr gets the index of the first non-file token in the tokarray. */
    tokptr= S[fileindex].filename; tokptr+= strlen(tokptr)+1;
    tokindex=0;
    for (; (*tokptr != ENDLIST) && (*tokptr != FILENAME); tokptr++) {

      /* If skip_cppstuff is on, ignore from a # up to
       * but not including the LINE token.
       */
      if (skip_cppstuff) {
        if ((*tokptr>=HASHdefine) && (*tokptr<=HASHwarning)) inside_cppstuff=1;
        if (*tokptr==LINE) inside_cppstuff=0;
	if (inside_cppstuff==1) { 
          switch(*tokptr) {
            case IDENTIFIER:
            case INTVAL:
            case LABEL:
            case CHARCONST:
            case STRINGLIT:
		tokptr+=2;
	  }
	  continue;
	}
      }

      /* Extract the unique id for tokens with values */
      switch(*tokptr) {
        case IDENTIFIER:
        case INTVAL:
        case LABEL:
        case CHARCONST:
        case STRINGLIT:
	  value= (*(tokptr+1) * 256) + *(tokptr+2);
	  S[fileindex].token[tokindex].id= value;
	  S[fileindex].token[tokindex].type=*tokptr;
	  tokptr+=2; break;
        default:
	  S[fileindex].token[tokindex].type=*tokptr;
      }
      tokindex++;
    }

    /* Add an ENDLIST token to the end */
    S[fileindex].token[tokindex].type= ENDLIST;
  }

  return (S);
}

/*
 * Print out lines of two files side-by-side. This is best viewed with a
 * terminal window 160 characters wide!
 */
void
paste_files(char *file1, int start1, int num1,
            char *file2, int start2, int num2)
{
  FILE *f1in, *f2in;
  char buf[1024];
  int i, maxlines;

  f1in = fopen(file1, "r");
  if (f1in == NULL)
    return;
  f2in = fopen(file2, "r");
  if (f2in == NULL) {
    fclose(f1in);
    return;
  }
  /* Get maximum number of lines */
  maxlines = (num1 > num2) ? num1 : num2;
  start1--;
  start2--;

  /* Read up to line start1 in file1 */
  while (start1--)
    fgets(buf, 1020, f1in);

  /* Read up to line start2 in file2 */
  while (start2--)
    fgets(buf, 1020, f2in);

  /* Now print out the lines */
  for (i = 0; i < maxlines; i++, num1--, num2--) {
    if (num1 <= 0)
      printf("%80s", " ");
    else {
      fgets(buf, 1020, f1in);
      buf[strlen(buf) - 1] = '\0';	/* Remove newline */
      printf("%-80s", buf);
    }
    printf("  ");
    if (num2 > 0) {
      fgets(buf, 1020, f2in);
      fputs(buf, stdout);
    }
  }
  printf("\n");
  if (num1>num2) printf("\n");
  fclose(f1in);
  fclose(f2in);
}


#ifdef ISOMORPH_CODE
/* Here are some comments on the isomorph[] arrays. We would like to
 * identify code fragments that are similar, even if variables have
 * been renamed or even declared in a different order. For example,
 * the following two functions are essentially identical.
 *
 * int maxofthree(int x, int y, int z)    int bigtriple(int b, int a, int c)
 * {                                      {
 *   if ((x>y) && (x>z)) return(x);          if ((a>b) && (a>c)) return(a);
 *   if (y>z) return(y);                     if (b>c) return(b);
 *   return(z);                              return(c);
 * }                                      }
 *
 * The tokenised forms of these functions are obviously different
 *
 * int id0(int id1,int id2,int id3)         int id0(int id1,int id2,int id3)
 * {                                        {
 *  if ((id1>id2)&&(id1>id3))                if ((id2>id1)&&(id2>id3))
 *    return(id1);                             return(id2);
 *  if(id2>id3) return(id2);                 if (id1>id3) return(id1);
 *  return(id3);                             return(id3);
 * }                                        }
 *
 * However, it is possible to compare these and show that they are identical.
 * If we start with the first "if" statement, a straight lexical comparison
 * shows that they are different. But at the first difference (id1 cf. id2),
 * we put an entry into the isoid[] array so that isoid[2]= 1, i.e. the
 * identifier #2 on the RHS is isomorphic to identifier #1 on the LHS.
 * Thus when we get to the first "return" statements, we can see that they
 * are also isomorphic. We can thus conclude that everything from the "if"
 * down to the "}" is isomorphic and thus identical.
 *
 * The other isomorph[] arrays are used for the other tokens with values:
 * isolabel[], isoint[], isochar[], isostr[].
 */

/* isomorphic_compare(): compare two runs of identical tokens
 * to see if they are also isomorphic w.r.t the token ids.
 * Returns 1 if they are, 0 if not.
 */
int
isomorphic_compare(struct token *s, struct token *d, struct token *send)
{
  
  unsigned short isoid[65536], isolabel[65536];
  unsigned short isoint[65536], isochar[65536];
  unsigned short isostr[65536];

  /* Clear the isomorph tables for this run */
  memset(isoid, 0, sizeof(isoid));
  memset(isolabel, 0, sizeof(isolabel));
  memset(isoint, 0, sizeof(isoint));
  memset(isochar, 0, sizeof(isochar));
  memset(isostr, 0, sizeof(isostr));

  /* Walk the array from start to end */
  while (1) {
    if (s==send) break;
    if (s->type == LINE) { s++; continue; }
    if (d->type == LINE) { d++; continue; }

    if (s->type!=d->type) { printf("Assert error! token mismatch!!\n"); }

    switch(s->type) {
      case IDENTIFIER:
	if ((isoid[d->id]>0) && (isoid[d->id]!=s->id)) return(0);
	if (isoid[d->id]==0) isoid[d->id]=s->id;
	break;
      case INTVAL:
	if ((isoint[d->id]>0) && (isoint[d->id]!=s->id)) return(0);
	if (isoint[d->id]==0) isoint[d->id]=s->id;
	break;
      case LABEL:
	if ((isolabel[d->id]>0) && (isolabel[d->id]!=s->id)) return(0);
	if (isolabel[d->id]==0) isolabel[d->id]=s->id;
	break;
      case CHARCONST:
	if ((isochar[d->id]>0) && (isochar[d->id]!=s->id)) return(0);
	if (isochar[d->id]==0) isochar[d->id]=s->id;
	break;
      case STRINGLIT:
	if ((isostr[d->id]>0) && (isostr[d->id]!=s->id)) return(0);
	if (isostr[d->id]==0) isostr[d->id]=s->id;
	break;
    }

    s++; d++;
  }
  return(1);
}
#endif

/*
 * compare_files():
 * This is the function that actually compares
 * tokens runs between two files, and prints out
 * runs bigger than a certain threshold.
 * This is slow and ugly brute-force code.
 * There must be a better way to do this.
 * Please let me know if there is!
 */
void
compare_files(int srcfileindex, int dstfileindex)
{
  /*
   * Stok, dtok, slinenum and dlinenum indicate the start of each run. s, d
   * and runcount are used inside the run.
   */
  struct token *stok, *dtok, *s, *d;	/* Pointers into tokarrays */
  struct token *tptr;
  int slinenum, dlinenum;
  int slinecount, dlinecount;
  int runcount;				/* # of tokens that matched */
#ifdef OPT_OVERLAP
  char *overlap;
  int srcindex, dstindex;

  /*
   * Build a "2D" array of overlap flags, so we can see where the overlaps
   * between the two files occur.
   */
  overlap = (char *)calloc(Srctree[srcfileindex].numtokens *
			 Dsttree[dstfileindex].numtokens * sizeof(char), 1);
  if (overlap == NULL)
    return;
#endif

  /* Do each token in this file in turn */
  slinenum = 1;
  for (stok = Srctree[srcfileindex].token; stok->type != ENDLIST; stok++) {
    if (stok->type == LINE) { slinenum++; continue; }

#ifdef DEBUG
    printf("Doing %s:%d token %s against %s\n",
	   Srctree[srcfileindex].filename, slinenum,
	   tokstring[stok->type], Dsttree[dstfileindex].filename);
#endif

    /* Try with every token in the dst file */
    dlinenum = 1;
    for (dtok = Dsttree[dstfileindex].token; dtok->type != ENDLIST; dtok++) {
#ifdef DEBUG
      printf("Doing dtok %d type %s\n",dtok, tokstring[dtok->type]);
#endif
      if (dtok->type == LINE) { dlinenum++; continue; }

      /* Start counting tokens in common */
      s = stok; d = dtok;
      slinecount = 0; dlinecount = 0;
      runcount = 0;

#ifdef OPT_OVERLAP
      /* Skip this comparison if we have already found a match */
      srcindex = s - Srctree[srcfileindex].token;
      dstindex = d - Dsttree[dstfileindex].token;
      /* printf("Checking ov[%d][%d]\n", srcindex, dstindex); */
      if (overlap[srcindex * Dsttree[dstfileindex].numtokens + dstindex] == 1)
	continue;
#endif

      while (1) {
	if ((s->type == ENDLIST) || (d->type == ENDLIST)) break;
	if (s->type == LINE) {
	  s++; slinecount++;
#ifdef OPT_OVERLAP
	  srcindex++;
#endif
	  continue;
	}
	if (d->type == LINE) {
	  d++; dlinecount++;
#ifdef OPT_OVERLAP
	  dstindex++;
#endif
	  continue;
	}
	if (s->type != d->type) break;
	runcount++;
#ifdef OPT_OVERLAP
	overlap[srcindex * Dsttree[dstfileindex].numtokens + dstindex] = 1;
	/* printf("Set ov[%d][%d]\n", srcindex, dstindex); */
	srcindex++; dstindex++;
#endif
	s++; d++;
      }

      /* Print out any run that meets the threshold */
      if (runcount >= threshold) {
#ifdef ISOMORPH_CODE
        /* Check this run of tokens for isomorphic correctness.
	 * If the call returns 0, the code doesn't match.
	 */
	if (isomorphic_compare(stok, dtok, s)==0) break;
#endif

	printf("%d  %s:%d  %s:%d\n", runcount,
	       Srctree[srcfileindex].filename, slinenum,
	       Dsttree[dstfileindex].filename, dlinenum);

	if (showtokens) {
	  /* Now print out the list of tokens */
	  for (tptr = stok; tptr < s; tptr++) {
	    fputs(tokstring[tptr->type], stdout);
	    switch(tptr->type) {
      	      case IDENTIFIER:
      	      case INTVAL:
        	printf("%d ", tptr->id); break;
      	      case LABEL:
        	printf("%d: ", tptr->id); break;
      	      case CHARCONST:
        	printf("%d' ", tptr->id); break;
      	      case STRINGLIT:
        	printf("%d\" ", tptr->id); break;
	    }
	  }
	  fputc('\n', stdout);
	  fputc('\n', stdout);
	}
	if (crosscompare)
	  paste_files(Srctree[srcfileindex].filename, slinenum, slinecount + 1,
		  Dsttree[dstfileindex].filename, dlinenum, dlinecount + 1);
      }
    }
  }
#ifdef OPT_OVERLAP
  free(overlap);
#endif
}

/*
 * Compare the two source trees. This is O(n2) which sucks.
 * Any improvements most welcome!
 */

void
compare_trees(void)
{
  int srcfindex, dstfindex;

  /* Walk the first list of filenames */
  for (srcfindex = 0; Srctree[srcfindex].filename != NULL; srcfindex++) {
    if (progress)
      fprintf(stderr, "Processing src file %s\n", Srctree[srcfindex].filename);

    /* Compare this file to the other files */
    for (dstfindex = 0; Dsttree[dstfindex].filename != NULL; dstfindex++) {
      compare_files(srcfindex, dstfindex);
    }
  }
}

void
usage(void)
{
 fprintf(stderr,
       "Usage: ctcompare [-p] [-t] [-x] [-i] [-n nnn] cft_file ctf_file\n");
 exit(1);
}


int
main(int argc, char *argv[])
{
  int ch;

  /* Process options */
  while ((ch = getopt(argc, argv, "ptxin:")) != -1) {
    switch (ch) {
    case 'p': progress = 1; break;
    case 'i': skip_cppstuff = 0; break;
    case 't':
      if (crosscompare) {
	fprintf(stderr, "-t and -x are mutually exclusive\n");
	crosscompare = 0;
      }
      showtokens = 1; break;
    case 'x':
      if (showtokens) {
	fprintf(stderr, "-t and -x are mutually exclusive\n");
	showtokens = 0;
      }
      crosscompare = 1; break;
    case 'n':
      threshold = atoi(optarg);
      if (threshold <= 0) {
	fprintf(stderr, "Bad value for -n\n");
	threshold = 10;
      }
      break;
    default: usage();
    }
  }
  argc -= optind; argv += optind;

  if (argc != 2) usage();

  Srctoklist = load_ctffile(argv[0]);
  Srctree = build_srcarray(Srctoklist);
  Dsttoklist = load_ctffile(argv[1]);
  Dsttree = build_srcarray(Dsttoklist);
  compare_trees();
  exit(0);
}
